@@ -1,42 +1,10 @@
-# This file is part of QuTiP: Quantum Toolbox in Python.
-#
-#    Copyright (c) 2011 and later, Paul D. Nation and Robert J. Johansson.
-#    All rights reserved.
-#
-#    Redistribution and use in source and binary forms, with or without
-#    modification, are permitted provided that the following conditions are
-#    met:
-#
-#    1. Redistributions of source code must retain the above copyright notice,
-#       this list of conditions and the following disclaimer.
-#
-#    2. Redistributions in binary form must reproduce the above copyright
-#       notice, this list of conditions and the following disclaimer in the
-#       documentation and/or other materials provided with the distribution.
-#
-#    3. Neither the name of the QuTiP: Quantum Toolbox in Python nor the names
-#       of its contributors may be used to endorse or promote products derived
-#       from this software without specific prior written permission.
-#
-#    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
-#    "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
-#    LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
-#    PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
-#    HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
-#    SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
-#    LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
-#    DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-#    THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
-#    (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
-#    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
-###############################################################################
-
 __all__ = ['Bloch']
 
 import os
 
 from numpy import (ndarray, array, linspace, pi, outer, cos, sin, ones, size,
                    sqrt, real, mod, append, ceil, arange)
+import numpy as np
 
 from packaging.version import parse as parse_version
 
@@ -73,12 +41,19 @@ try:
 
             self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
             FancyArrowPatch.draw(self, renderer)
-except:
+
+        def do_3d_projection(self, renderer=None):
+            # only called by matplotlib >= 3.5
+            xs3d, ys3d, zs3d = self._verts3d
+            xs, ys, zs = proj3d.proj_transform(xs3d, ys3d, zs3d, self.axes.M)
+            self.set_positions((xs[0], ys[0]), (xs[1], ys[1]))
+            return np.min(zs)
+except ImportError:
     pass
 
 try:
     from IPython.display import display
-except:
+except ImportError:
     pass
 
 
@@ -145,6 +120,7 @@ class Bloch:
                  background=False):
         # Figure and axes
         self.fig = fig
+        self._ext_fig = fig is not None
         self.axes = axes
         # Background axes, default = False
         self.background = background
@@ -204,20 +180,20 @@ class Bloch:
         self.points = []
         # Data for Bloch vectors
         self.vectors = []
+        # Transparency of vectors, alpha value from 0 to 1
+        self.vector_alpha = []
         # Data for annotations
         self.annotations = []
         # Number of times sphere has been saved
         self.savenum = 0
         # Style of points, 'm' for multiple colors, 's' for single color
         self.point_style = []
-
-        # status of rendering
-        self._rendered = False
-        # status of showing
-        if fig is None:
-            self._shown = False
-        else:
-            self._shown = True
+        # Transparency of points, alpha value from 0 to 1
+        self.point_alpha = []
+        # Data for line segment
+        self._lines = []
+        # Data for arcs and arc style
+        self._arcs = []
 
     def set_label_convention(self, convention):
         """Set x, y and z labels according to one of conventions.
@@ -233,9 +209,9 @@ class Bloch:
             - "01"
             - "polarization jones"
             - "polarization jones letters"
-              see also: http://en.wikipedia.org/wiki/Jones_calculus
+              see also: https://en.wikipedia.org/wiki/Jones_calculus
             - "polarization stokes"
-              see also: http://en.wikipedia.org/wiki/Stokes_parameters
+              see also: https://en.wikipedia.org/wiki/Stokes_parameters
         """
         ketex = "$\\left.|%s\\right\\rangle$"
         # \left.| is on purpose, so that every ket has the same size
@@ -327,9 +303,13 @@ class Bloch:
         self.points = []
         self.vectors = []
         self.point_style = []
+        self.point_alpha = []
+        self.vector_alpha = []
         self.annotations = []
+        self._lines = []
+        self._arcs = []
 
-    def add_points(self, points, meth='s'):
+    def add_points(self, points, meth='s', alpha=1.0):
         """Add a list of data points to bloch sphere.
 
         Parameters
@@ -340,8 +320,18 @@ class Bloch:
         meth : {'s', 'm', 'l'}
             Type of points to plot, use 'm' for multicolored, 'l' for points
             connected with a line.
+
+        alpha : float, default=1.
+            Transparency value for the vectors. Values between 0 and 1.
+
+        .. note::
+
+           When using ``meth=l`` in QuTiP 4.6, the line transparency defaulted
+           to ``0.75`` and there was no way to alter it.
+           When the ``alpha`` parameter was added in QuTiP 4.7, the default
+           became ``alpha=1.0`` for values of ``meth``.
         """
-        if not isinstance(points[0], (list, ndarray)):
+        if not isinstance(points[0], (list, tuple, ndarray)):
             points = [[points[0]], [points[1]], [points[2]]]
         points = array(points)
         if meth == 's':
@@ -352,14 +342,17 @@ class Bloch:
                 pnts = points
             self.points.append(pnts)
             self.point_style.append('s')
+            self.point_alpha.append(alpha)
         elif meth == 'l':
             self.points.append(points)
             self.point_style.append('l')
+            self.point_alpha.append(alpha)
         else:
             self.points.append(points)
             self.point_style.append('m')
+            self.point_alpha.append(alpha)
 
-    def add_states(self, state, kind='vector'):
+    def add_states(self, state, kind='vector', alpha=1.0):
         """Add a state vector Qobj to Bloch sphere.
 
         Parameters
@@ -369,6 +362,9 @@ class Bloch:
 
         kind : {'vector', 'point'}
             Type of object to plot.
+
+        alpha : float, default=1.
+            Transparency value for the vectors. Values between 0 and 1.
         """
         if isinstance(state, Qobj):
             state = [state]
@@ -379,23 +375,28 @@ class Bloch:
                    expect(sigmaz(), st)]
 
             if kind == 'vector':
-                self.add_vectors(vec)
+                self.add_vectors(vec, alpha=alpha)
             elif kind == 'point':
-                self.add_points(vec)
+                self.add_points(vec, alpha=alpha)
 
-    def add_vectors(self, vectors):
+    def add_vectors(self, vectors, alpha=1.0):
         """Add a list of vectors to Bloch sphere.
 
         Parameters
         ----------
         vectors : array_like
             Array with vectors of unit length or smaller.
+
+        alpha : float, default=1.
+            Transparency value for the vectors. Values between 0 and 1.
         """
-        if isinstance(vectors[0], (list, ndarray)):
+        if isinstance(vectors[0], (list, tuple, ndarray)):
             for vec in vectors:
                 self.vectors.append(vec)
+                self.vector_alpha.append(alpha)
         else:
             self.vectors.append(vectors)
+            self.vector_alpha.append(alpha)
 
     def add_annotation(self, state_or_vector, text, **kwargs):
         """
@@ -434,11 +435,135 @@ class Bloch:
                                  'text': text,
                                  'opts': kwargs})
 
+    def add_arc(self, start, end, fmt="b", steps=None, **kwargs):
+        """Adds an arc between two points on a sphere. The arc is set to be
+        blue solid curve by default.
+
+        The start and end points must be on the same sphere (i.e. have the
+        same radius) but need not be on the unit sphere.
+
+        Parameters
+        ----------
+        start : Qobj or array-like
+            Array with cartesian coordinates of the first point, or a state
+            vector or density matrix that can be mapped to a point on or
+            within the Bloch sphere.
+        end : Qobj or array-like
+            Array with cartesian coordinates of the second point, or a state
+            vector or density matrix that can be mapped to a point on or
+            within the Bloch sphere.
+        fmt : str, default: "b"
+            A matplotlib format string for rendering the arc.
+        steps : int, default: None
+            The number of segments to use when rendering the arc. The default
+            uses 100 steps times the distance between the start and end points,
+            with a minimum of 2 steps.
+        **kwargs : dict
+            Additional parameters to pass to the matplotlib .plot function
+            when rendering this arc.
+        """
+        if isinstance(start, Qobj):
+            pt1 = [
+                expect(sigmax(), start),
+                expect(sigmay(), start),
+                expect(sigmaz(), start),
+            ]
+        else:
+            pt1 = start
+
+        if isinstance(end, Qobj):
+            pt2 = [
+                expect(sigmax(), end),
+                expect(sigmay(), end),
+                expect(sigmaz(), end),
+            ]
+        else:
+            pt2 = end
+
+        pt1 = np.asarray(pt1)
+        pt2 = np.asarray(pt2)
+
+        len1 = np.linalg.norm(pt1)
+        len2 = np.linalg.norm(pt2)
+        if len1 < 1e-12 or len2 < 1e-12:
+            raise ValueError('Polar and azimuthal angles undefined at origin.')
+        elif abs(len1 - len2) > 1e-12:
+            raise ValueError("Points not on the same sphere.")
+        elif (pt1 == pt2).all():
+            raise ValueError(
+                "Start and end represent the same point. No arc can be formed."
+            )
+        elif (pt1 == -pt2).all():
+            raise ValueError(
+                "Start and end are diagonally opposite, no unique arc is"
+                " possible."
+            )
+
+        if steps is None:
+            steps = int(np.linalg.norm(pt1 - pt2) * 100)
+            steps = max(2, steps)
+        t = np.linspace(0, 1, steps)
+        # All the points in this line are contained in the plane defined
+        # by pt1, pt2 and the origin.
+        line = pt1[:, np.newaxis] * t + pt2[:, np.newaxis] * (1 - t)
+        # Normalize all the points in the line so that are distance len1 from
+        # the origin.
+        arc = line * len1 / np.linalg.norm(line, axis=0)
+        self._arcs.append([arc, fmt, kwargs])
+
+    def add_line(self, start, end, fmt="k", **kwargs):
+        """Adds a line segment connecting two points on the bloch sphere.
+
+        The line segment is set to be a black solid line by default.
+
+        Parameters
+        ----------
+        start : Qobj or array-like
+            Array with cartesian coordinates of the first point, or a state
+            vector or density matrix that can be mapped to a point on or
+            within the Bloch sphere.
+        end : Qobj or array-like
+            Array with cartesian coordinates of the second point, or a state
+            vector or density matrix that can be mapped to a point on or
+            within the Bloch sphere.
+        fmt : str, default: "k"
+            A matplotlib format string for rendering the line.
+        **kwargs : dict
+            Additional parameters to pass to the matplotlib .plot function
+            when rendering this line.
+        """
+        if isinstance(start, Qobj):
+            pt1 = [
+                expect(sigmax(), start),
+                expect(sigmay(), start),
+                expect(sigmaz(), start),
+            ]
+        else:
+            pt1 = start
+
+        if isinstance(end, Qobj):
+            pt2 = [
+                expect(sigmax(), end),
+                expect(sigmay(), end),
+                expect(sigmaz(), end),
+            ]
+        else:
+            pt2 = end
+
+        pt1 = np.asarray(pt1)
+        pt2 = np.asarray(pt2)
+
+        x = [pt1[1], pt2[1]]
+        y = [-pt1[0], -pt2[0]]
+        z = [pt1[2], pt2[2]]
+        v = [x, y, z]
+        self._lines.append([v, fmt, kwargs])
+
     def make_sphere(self):
         """
         Plots Bloch sphere and data sets.
         """
-        self.render(self.fig, self.axes)
+        self.render()
 
     def run_from_ipython(self):
         try:
@@ -447,30 +572,50 @@ class Bloch:
         except NameError:
             return False
 
-    def render(self, fig=None, axes=None):
+    def _is_inline_backend(self):
+        backend = matplotlib.get_backend()
+        return backend == "module://matplotlib_inline.backend_inline"
+
+    def render(self):
         """
         Render the Bloch sphere and its data sets in on given figure and axes.
         """
-        if self._rendered:
-            self.axes.clear()
+        if not self._ext_fig and not self._is_inline_backend():
+            # If no external figure was supplied, we check to see if the
+            # figure we created in a previous call to .render() has been
+            # closed, and re-create if has been. This has the unfortunate
+            # side effect of losing any modifications made to the axes or
+            # figure, but the alternative is to crash the matplotlib backend.
+            #
+            # The inline backend used by, e.g. jupyter notebooks, is happy to
+            # use closed figures so we leave those figures intact.
+            if (
+                self.fig is not None and
+                not plt.fignum_exists(self.fig.number)
+            ):
+                self.fig = None
+                self.axes = None
 
-        self._rendered = True
-
-        # Figure instance for Bloch sphere plot
-        if not fig:
+        if self.fig is None:
             self.fig = plt.figure(figsize=self.figsize)
+            if self._is_inline_backend():
+                # We immediately close the inline figure do avoid displaying
+                # the figure twice when .show() calls display.
+                plt.close(self.fig)
 
-        if not axes:
-            self.axes = _axes3D(self.fig, azim=self.view[0],
-                                elev=self.view[1])
+        if self.axes is None:
+            self.axes = _axes3D(self.fig, azim=self.view[0], elev=self.view[1])
 
+        # Clearing the axes is horrifically slow and loses a lot of the
+        # axes state, but matplotlib doesn't seem to provide a better way
+        # to redraw Axes3D. :/
+        self.axes.clear()
+        self.axes.grid(False)
         if self.background:
-            self.axes.clear()
             self.axes.set_xlim3d(-1.3, 1.3)
             self.axes.set_ylim3d(-1.3, 1.3)
             self.axes.set_zlim3d(-1.3, 1.3)
         else:
-            self.plot_axes()
             self.axes.set_axis_off()
             self.axes.set_xlim3d(-0.7, 0.7)
             self.axes.set_ylim3d(-0.7, 0.7)
@@ -479,14 +624,19 @@ class Bloch:
         # Matplotlib did this stretching for < 3.3.0, but not above.
         if parse_version(matplotlib.__version__) >= parse_version('3.3'):
             self.axes.set_box_aspect((1, 1, 1))
+        if not self.background:
+            self.plot_axes()
 
-        self.axes.grid(False)
         self.plot_back()
         self.plot_points()
         self.plot_vectors()
+        self.plot_lines()
+        self.plot_arcs()
         self.plot_front()
         self.plot_axes_labels()
         self.plot_annotations()
+        # Trigger an update of the Bloch sphere if it is already shown:
+        self.fig.canvas.draw()
 
     def plot_back(self):
         # back half of sphere
@@ -555,14 +705,14 @@ class Bloch:
         self.axes.text(0, 0, self.zlpos[0], self.zlabel[0], **opts)
         self.axes.text(0, 0, self.zlpos[1], self.zlabel[1], **opts)
 
-        for a in (self.axes.w_xaxis.get_ticklines() +
-                  self.axes.w_xaxis.get_ticklabels()):
+        for a in (self.axes.xaxis.get_ticklines() +
+                  self.axes.xaxis.get_ticklabels()):
             a.set_visible(False)
-        for a in (self.axes.w_yaxis.get_ticklines() +
-                  self.axes.w_yaxis.get_ticklabels()):
+        for a in (self.axes.yaxis.get_ticklines() +
+                  self.axes.yaxis.get_ticklabels()):
             a.set_visible(False)
-        for a in (self.axes.w_zaxis.get_ticklines() +
-                  self.axes.w_zaxis.get_ticklabels()):
+        for a in (self.axes.zaxis.get_ticklines() +
+                  self.axes.zaxis.get_ticklabels()):
             a.set_visible(False)
 
     def plot_vectors(self):
@@ -574,19 +724,21 @@ class Bloch:
             zs3d = self.vectors[k][2] * array([0, 1])
 
             color = self.vector_color[mod(k, len(self.vector_color))]
+            alpha = self.vector_alpha[k]
 
             if self.vector_style == '':
                 # simple line style
                 self.axes.plot(xs3d, ys3d, zs3d,
                                zs=0, zdir='z', label='Z',
-                               lw=self.vector_width, color=color)
+                               lw=self.vector_width, color=color,
+                               alpha=alpha)
             else:
                 # decorated style, with arrow heads
                 a = Arrow3D(xs3d, ys3d, zs3d,
                             mutation_scale=self.vector_mutation,
                             lw=self.vector_width,
                             arrowstyle=self.vector_style,
-                            color=color)
+                            color=color, alpha=alpha)
 
                 self.axes.add_artist(a)
 
@@ -611,7 +763,7 @@ class Bloch:
                     - real(self.points[k][0][indperm]),
                     real(self.points[k][2][indperm]),
                     s=self.point_size[mod(k, len(self.point_size))],
-                    alpha=1,
+                    alpha=self.point_alpha[k],
                     edgecolor=None,
                     zdir='z',
                     color=self.point_color[mod(k, len(self.point_color))],
@@ -619,7 +771,8 @@ class Bloch:
 
             elif self.point_style[k] == 'm':
                 pnt_colors = array(self.point_color *
-                                   int(ceil(num / float(len(self.point_color)))))
+                                   int(ceil(num / float(
+                                       len(self.point_color)))))
 
                 pnt_colors = pnt_colors[0:num]
                 pnt_colors = list(pnt_colors[indperm])
@@ -628,17 +781,17 @@ class Bloch:
                 self.axes.scatter(real(self.points[k][1][indperm]),
                                   -real(self.points[k][0][indperm]),
                                   real(self.points[k][2][indperm]),
-                                  s=s, alpha=1, edgecolor=None,
-                                  zdir='z', color=pnt_colors,
-                                  marker=marker)
+                                  s=s, alpha=self.point_alpha[k],
+                                  edgecolor=None, zdir='z',
+                                  color=pnt_colors, marker=marker)
 
             elif self.point_style[k] == 'l':
                 color = self.point_color[mod(k, len(self.point_color))]
                 self.axes.plot(real(self.points[k][1]),
                                -real(self.points[k][0]),
                                real(self.points[k][2]),
-                               alpha=0.75, zdir='z',
-                               color=color)
+                               alpha=self.point_alpha[k],
+                               zdir='z', color=color)
 
     def plot_annotations(self):
         # -X and Y data are switched for plotting purposes
@@ -652,17 +805,34 @@ class Bloch:
             self.axes.text(vec[1], -vec[0], vec[2],
                            annotation['text'], **opts)
 
+    def plot_lines(self):
+        for line, fmt, kw in self._lines:
+            self.axes.plot(line[0], line[1], line[2], fmt, **kw)
+
+    def plot_arcs(self):
+        for arc, fmt, kw in self._arcs:
+            self.axes.plot(arc[1, :], -arc[0, :], arc[2, :], fmt, **kw)
+
     def show(self):
         """
         Display Bloch sphere and corresponding data sets.
+
+        Notes
+        -----
+
+        When using inline plotting in Jupyter notebooks, any figure created
+        in a notebook cell is displayed after the cell executes. Thus if you
+        create a figure yourself and use it create a Bloch sphere with
+        ``b = Bloch(..., fig=fig)`` and then call ``b.show()`` in the same
+        cell, then the figure will be displayed twice. If you do create your
+        own figure, the simplest solution to this is to not call ``.show()``
+        in the cell you create the figure in.
         """
-        self.render(self.fig, self.axes)
+        self.render()
         if self.run_from_ipython():
-            if self._shown:
-                display(self.fig)
+            display(self.fig)
         else:
             self.fig.show()
-        self._shown = True
 
     def save(self, name=None, format='png', dirc=None, dpin=None):
         """Saves Bloch sphere to file of type ``format`` in directory ``dirc``.
@@ -686,7 +856,7 @@ class Bloch:
         File containing plot of Bloch sphere.
 
         """
-        self.render(self.fig, self.axes)
+        self.render()
         # Conditional variable for first argument to savefig
         # that is set in subsequent if-elses
         complete_path = ""
